from __future__ import annotations

import logging

import toga
from toga.style import Pack
from toga.style.pack import CENTER, ROW  # type: ignore

from pyMOSF.core import Event, EventType, ServiceRegistry
from pyMOSF.toga import TogaComponent, TogaMultiLayoutApp, TogaStackedLayout
from pyMOSF.toga.services import FileOpen

log = logging.getLogger(__name__)


class ToolbarComponent(TogaComponent):
    def __init__(self, layout: TogaStackedLayout, **kwargs) -> None:
        icon_style = Pack(width=120,
                          height=46,
                          padding=2)
        super().__init__(layout, style=Pack(
            direction=ROW, alignment=CENTER, padding=1),
            children=[
                toga.Button(text="press me",
                            style=icon_style,
                            on_press=self.button_pressed),
                toga.Button(text="Open file",
                            id="open_file",
                            style=icon_style),
        ])

    @property
    def parent_layout(self) -> ExampleViewLayout:
        return self._layout  # type: ignore

    def button_pressed(self, widget):
        self.parent_layout.button_pressed()  # type: ignore

    def file_selected(self, paths):
        self.parent_layout.file_opened(paths.file)  # type: ignore

    def _non_ios_config(self):
        ServiceRegistry().bind_event(
            Event("open_file",
                  EventType.ON_PRESS,
                  FileOpen(dialog_title="Open file",
                           multiple_select=False),
                  service_callback=self.file_selected))

    def on_linux_config(self):
        self._non_ios_config()

    def on_windows_config(self):
        self._non_ios_config()

    def on_darwin_config(self):
        self._non_ios_config()

    def _ios_config(self):
        # Load the library inside the functoin
        # to avoid dependency problem on other platform
        from pyMOSF.services.open_file_ios import IOSFileOpen
        ServiceRegistry().bind_event(
            Event("open_file",
                  EventType.ON_PRESS,
                  IOSFileOpen(initial_directory=self.ml_app.data_path,
                              multiple_select=False),
                  service_callback=self.file_selected))

    def on_ios_config(self):
        self._ios_config()

    def on_ipados_config(self):
        self._ios_config()


class TextViewComponent(TogaComponent):

    def __init__(self, layout: TogaStackedLayout, **kwargs) -> None:
        self._label = toga.Label("", style=Pack(padding=10, flex=1))
        super().__init__(layout,
                         style=Pack(direction=ROW, alignment=CENTER,
                                    padding=1, flex=1),
                         children=[
                             self._label,
                         ])

    @property
    def label(self) -> toga.Label:
        return self._label


class ExampleViewLayout(TogaStackedLayout):
    def __init__(self, app: TogaMultiLayoutApp):
        super().__init__(app,
                         ToolbarComponent,
                         TextViewComponent)
        self.counter = 1

    @property
    def files_toolbar(self) -> ToolbarComponent:
        return self[ToolbarComponent]

    @property
    def label_view(self) -> toga.Label:
        if self[TextViewComponent] is None:
            raise ValueError("Main box has not been initialized.")
        return self[TextViewComponent]._label

    def on_common_config(self):
        ############
        # Settings
        self.settings = self.ml_app.settings

    def on_load(self):
        return super().on_load()

    def button_pressed(self):
        self.label_view.text = f"Button pressed {self.counter} times."
        self.counter += 1

    def file_opened(self, file: str):   # type: ignore
        self.label_view.text = f"File opened: \n{file}"




import logging

from .app import create_app as create_toga_app

log = logging.getLogger(__name__)

if __name__ == "__main__":
    log.info("main is called.")
    try:
        create_toga_app().main_loop()
    except Exception as e:
        print(e)
        log.error("Exception in main loop. \n"
                  f"Exception: {e}"
                  f"\nTraceback: {e.__traceback__}")



from __future__ import annotations

import logging

from kivy.uix.label import Label  # type: ignore

from pyMOSF.core import Event, EventType, ServiceRegistry
from pyMOSF.kivy import KivyComponent, KivyMultiLayoutApp, KivyStackedLayout
from pyMOSF.kivy.services.io import FileOpen

# from pathlib import Path


log = logging.getLogger(__name__)

TOOLBAR_HEIGHT = 125


# Builder.load_file(str(Path("layout.kv").resolve()))

class ToolbarComponent(KivyComponent):
    @property
    def label(self):
        return self.ids.label

    @property
    def parent_layout(self) -> ExampleViewLayout:
        return self._layout  # type: ignore

    def button_pressed(self):
        self.parent_layout.button_pressed()  # type: ignore

    def file_selected(self, paths):
        self.parent_layout.file_opened(paths.file)  # type: ignore

    def _non_ios_config(self):
        ServiceRegistry().bind_event(
            Event("open_file",
                  EventType.ON_PRESS,
                  FileOpen(dialog_title="Open file",
                           multiple_select=False),
                  service_callback=self.file_selected))

    def on_linux_config(self):
        self._non_ios_config()

    def on_windows_config(self):
        self._non_ios_config()

    def on_darwin_config(self):
        self._non_ios_config()

    def _ios_config(self):
        # Load the library inside the functoin
        # to avoid dependency problem on other platform
        from pyMOSF.services.open_file_ios import IOSFileOpen
        ServiceRegistry().bind_event(
            Event("open_file",
                  EventType.ON_PRESS,
                  IOSFileOpen(initial_directory=self.ml_app.data_path,
                              multiple_select=False),
                  service_callback=self.file_selected))

    def on_ios_config(self):
        self._ios_config()

    def on_ipados_config(self):
        self._ios_config()


class TextViewComponent(KivyComponent):

    @property
    def label(self) -> Label:
        return self.ids.label


class ExampleViewLayout(KivyStackedLayout):
    def __init__(self, app: KivyMultiLayoutApp):
        super().__init__(app,
                         ToolbarComponent,
                         TextViewComponent)

        self.counter = 1

    @property
    def files_toolbar(self) -> ToolbarComponent:
        return self[ToolbarComponent]

    @property
    def label_view(self) -> Label:
        if self[TextViewComponent] is None:
            raise ValueError("Main box has not been initialized.")
        return self[TextViewComponent].label

    def on_common_config(self):
        ############
        # Settings
        self.settings = self.ml_app.settings

    def on_load(self):
        return super().on_load()

    def button_pressed(self):
        self.label_view.text = f"Button pressed {self.counter} times."
        self.counter += 1

    def file_opened(self, file: str):   # type: ignore
        self.label_view.text = f"File opened: \n{file}"



import logging
from pathlib import Path
from typing import Any

import toga

import pyMOSF.core as core
from pyMOSF.config import Dict
from pyMOSF.core import safe_async_call

log = logging.getLogger(__name__)


class FileOpen(core.AsyncService):
    def __init__(self,
                 dialog_title: str = "Open file",
                 initial_directory: Path | str | None = None,
                 file_types: list[str] | None = None,
                 multiple_select: bool = False) -> None:
        super().__init__()
        self.dialog_title = dialog_title
        self.initial_directory = initial_directory
        self.file_types = file_types
        self.multiple_select = multiple_select

    @safe_async_call(log)
    async def handle_event(self,
                           widget: Any,
                           app,
                           service_callback=None,
                           *args, **kwargs):

        fname = await widget.app.dialog(
            toga.OpenFileDialog(title=self.dialog_title,
                                initial_directory=self.initial_directory,
                                file_types=self.file_types,
                                multiple_select=self.multiple_select))
        if fname is None:
            return  # No file selected
        if service_callback is not None:
            if self.multiple_select:
                # If multiple files are selected, return a list of file names
                service_callback(Dict(files=fname))
            else:
                service_callback(Dict(file=fname))
        return



"""The core module of toga related implimentation.

   Notes:
   ------
   Configurable
       1- All three AbstractApp (TogaMultiLayoutApp),
          AbstractLayout (TogaLayout) and TogaComponent
          are 'Configurable'. So, they can implement
          os dependent configuration.
        2- AbstractApp calls Configurable's _set_config once in its
           lifcycle:  on_begin() -- for TogaMultiLayoutApp
           it happens in startup.
        3- AbstractLayout (TogaLayout) calls 'Configurable' once on its
           first load: on_load -- for TogaLayout, it is called
           on 'show_layout' of the app.
        4- AbstractLayout (TogaLayout) are created once, but will be used agian
           in TogaMultiLayoutApp's 'show_layout'. So, their
           configuration will happens only once. However, the
           AbstractLayout's 'on_load' function will be call everytime
           that it is loaeded again.
        5- TogaComponent's 'Configurable' are called once by the
           time their layout 'Configurable' are called.

   TogaStackedLayout (TogaLayout)
       1- It can have one or more TogaComponents. It is just need
          the class names of the TogaComponents in its constructor.
       2- Each component is accessible by its class name in the
          layout: e.g. layout[TogaComponentClass].
       3- It is possible to directly access the component method by
          decorating them with '@component_method'. Aftre that, we
          can call them, e.g., like layout.component_function().
          If it is not decorated. it can riase 'ServiceArgumentError'
          exception when thery are used for services argument.

   TogaComponent
       1- They must implement all thier UI elements inside
          __init__ and pass it to its super argument, after their
          parent layout.
       2- To make any of its method directly accessible in their
          layout, we can use '@component_method' decorator.
          If it is not decorated. it can riase 'ServiceArgumentError'
          exception when they are used for services argument.
"""
from __future__ import annotations

import logging

import toga
from toga.style import Pack
from toga.style.pack import CENTER, COLUMN  # type: ignore

from pyMOSF.config import Configurable, GUIFramework, Settings
from pyMOSF.core import (
    Component,
    Layout,
    MultiLayoutApp,
    ServiceRegistry,
    StackedLayout,
)

ServiceRegistry._framework = GUIFramework.TOGA


log = logging.getLogger(__name__)


class Promiseable:
    def __init__(self, **kwargs) -> None:
        super().__init__()

    @property
    def ml_app(self) -> TogaMultiLayoutApp:
        if self._app is None:
            raise ValueError("app has not been initialized.")
        return self._app

    @ml_app.setter
    def ml_app(self, value: TogaMultiLayoutApp):
        self._app = value

    def promise(self, func):
        if not callable(func):
            raise ValueError("The argument must be callable. "
                             "(Turn it into a lambda if it's a statment.)")

        def wrapper():
            func()

        self.ml_app.loop.call_soon_threadsafe(wrapper)


class TogaLayout(Layout, Promiseable):
    def __init__(self, app: TogaMultiLayoutApp) -> None:
        self._app = app
        super().__init__(app)

    @property
    def ml_app(self) -> TogaMultiLayoutApp:
        if self._app is None:
            raise ValueError("app has not been initialized.")
        return self._app


class TogaComponent(Component, Configurable, toga.Box, Promiseable):
    def __init__(self, layout: TogaStackedLayout, **kwargs) -> None:
        self._layout = layout
        super().__init__(layout, **kwargs)

    @property
    def ml_app(self) -> TogaMultiLayoutApp:
        if self._layout.ml_app is None:
            raise ValueError("app has not been initialized.")
        return self._layout.ml_app

    @property
    def parent_layout(self) -> TogaStackedLayout:
        return self._layout


class TogaStackedLayout(StackedLayout, Promiseable):
    def __init__(self,
                 app: TogaMultiLayoutApp,
                 *types):  # type: ignore
        self._box = toga.Box(style=Pack(direction=COLUMN,
                                        alignment=CENTER, flex=1)
                             )
        self._app = app
        super().__init__(app, self._box, *types)

    @property
    def ml_app(self) -> TogaMultiLayoutApp:
        if self._app is None:
            raise ValueError("app has not been initialized.")
        return self._app

    @ml_app.setter
    def ml_app(self, value: TogaMultiLayoutApp):
        self._app = value

    def _add_to_main_box(self, box: toga.Box, element: toga.Box):
        box.add(element)


class TogaMultiLayoutApp(MultiLayoutApp, toga.App):
    def __init__(self,
                 init_layout: TogaLayout | TogaStackedLayout,
                 settings: Settings | None = None,
                 **kwargs) -> None:
        self._main_container = toga.Box(
            style=Pack(direction=COLUMN, alignment=CENTER, flex=1)
        )
        self._current_layout = init_layout
        super().__init__(self._main_container, init_layout, settings, **kwargs)

    @property
    def path(self):
        return self.paths.app  # type: ignore

    @property
    def data_path(self):
        return self.paths.data  # type: ignore

    @property
    def current_layout(self) -> TogaLayout | TogaStackedLayout:
        return self._current_layout

    def _add_layout(self, layout: TogaLayout | TogaStackedLayout):
        self._main_container.clear()
        self._main_container.add(layout.main_box)

    def startup(self):
        """Construct and show the Toga application (it's called by toga app).

        It also call the layout build and bind the serivices handlers
        to UI elements on_press call back, if their ids correspond to
        an id in ServiceRegistry.
        """
        self.commands.clear()
        # self.main_window = toga.MainWindow(title=self.formal_name)
        self.main_window: toga.Window = toga.Window()
        self.main_window.content = self._main_container
        # Call on_load in AbstractApp to register the event once
        self.on_begin()
        # It calls once
        self._add_layout(self._current_layout)
        #
        self.main_window.show()  # type: ignore

    def on_exit(self) -> bool:
        return super().on_end()



from abc import abstractmethod
from typing import Any

import pyMOSF.core as core
from pyMOSF.config import Dict
from pyMOSF.core.pipelines import Process


class ToOpenCVImageProcess(Process):
    @abstractmethod
    def __call__(self, *, image, **kwargs) -> Dict:
        pass


class ToFrameworkImageProcess(Process):
    @abstractmethod
    def __call__(self, *, image: Any, **kwargs) -> Dict:
        pass


class ProcessesRegistry(object):
    """A singleton object that store Framework specific Process.
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ProcessesRegistry, cls).__new__(cls)
            # Put any initialization here.
        return cls._instance
    processes = {}

    def __getitem__(self, key):
        return self.processes[key]

    def __setitem__(self, key, newvalue):
        self.processes[key] = newvalue


class SyncImageService(core.SyncService):
    def __init__(self) -> None:
        registry = ProcessesRegistry()
        self.toOpenCV = registry[ToOpenCVImageProcess]
        self.toFramework = registry[ToFrameworkImageProcess]


class AsyncImageService(core.AsyncService):
    def __init__(self) -> None:
        registry = ProcessesRegistry()
        self.toOpenCV = registry[ToOpenCVImageProcess]
        self.toFramework = registry[ToFrameworkImageProcess]



from .__boxed__ import Component, Layout, MultiLayoutApp, StackedLayout  # noqa
from .__core__ import (  # noqa; ServiceCallback,
    AbstractApp,
    AbstractLayout,
    AsyncService,
    Event,
    EventDispatcher,
    EventType,
    ServiceArgumentError,
    ServiceRegistry,
    SyncService,
)
from .__decorators__ import (  # noqa; service_callback,
    component_method,
    processFactory,
    processLogic,
    processLogicProperty,
    silence_crossed_events,
)
from .__image_services__ import ProcessesRegistry  # noqa
from .__image_services__ import ToFrameworkImageProcess as ToFrameworkImage  # noqa
from .__image_services__ import ToOpenCVImageProcess as ToOpenCVImage  # noqa
from .__loggers__ import __dummy__  # noqa
from .__safe_calls__ import int_, safe_async_call, safe_call  # noqa
from .pipelines import ProcessPassThrough  # noqa
from .pipelines import ImageProcessingPipeline, Process  # noqa



"""Configuartion/setting module.

   Notes:
   ------
       1- Configurations are readonly, and are stored in 'resource' folder.
       2- Setting are writable and are stored in users' data folder,
          dependening on the os.
"""
from __future__ import annotations

import json
import platform
from enum import Enum
from pathlib import Path

from addict import Dict as DefaultDict


class GUIFramework(str, Enum):
    TOGA = "TOGA"
    KIVY = "KIVY"
    UKNOWN = "UKNOWN"


class Dict(DefaultDict):
    def __missing__(self, key) -> None:
        # raise KeyError(key)
        # calling dict.unassinged properties return None
        return None


class _Dict(DefaultDict):
    def __missing__(self, key):
        raise KeyError(key)


class Settings(DefaultDict):
    __instance = None
    __is_changed = False
    __loading_path = ""
    FILE_NAME = "settings.json"

    def __init__(__self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def load(data_path, default_conf={}) -> Settings:
        if Settings.__instance is None:
            try:

                with open(data_path / Settings.FILE_NAME, "r") as f:
                    __setting = json.load(f)
                    Settings.__loading_path = str(
                        data_path / Settings.FILE_NAME)

                    # Todo:Update thye default setting if new version
                    # has new entry.

                    Settings.__instance = Settings(**__setting)
                    # Check if the settings are empty
                    # and set the default settings

            except FileNotFoundError:
                Path(data_path).mkdir(parents=True, exist_ok=True)
                ###################################
                # Define the default settings here
                with open(data_path / Settings.FILE_NAME, "w") as f:
                    json.dump(default_conf, f, indent=4)
                    Settings.__loading_path = str(
                        data_path / Settings.FILE_NAME)
                Settings.__instance = Settings(**default_conf)
                Settings.__is_changed = False

        return Settings.__instance

    def __setitem__(self, name, value):
        Settings.__is_changed = True
        return super().__setitem__(name, value)

    def save(self):
        with open(Settings.__loading_path, "w") as f:
            json.dump(self, f, indent=4)
            Settings.__is_changed = False

    def on_end(self):
        if Settings.__is_changed:
            self.save()


class Configurable:
    def on_common_config(self):
        pass

    def on_linux_config(self):
        pass

    def on_darwin_config(self):
        pass

    def on_ios_config(self):
        pass

    def on_ipados_config(self):
        pass

    def on_windows_config(self):
        pass

    def on_others_config(self):
        raise NotImplementedError()

    def _set_config(self):
        """Layout related config (e.g. bindings).
        """
        self.on_common_config()
        os = platform.system().lower()
        match os:
            case "linux":
                self.on_linux_config()
            case "darwin":
                self.on_darwin_config()
            case "ios":
                self.on_ios_config()
            case "ipados":
                self.on_ipados_config()
            case "windows":
                self.on_windows_config()
            case _:
                self.on_others_config()



import pytest

from pyMOSF.core import processFactory, processLogic, processLogicProperty
from pyMOSF.core.pipelines import (
    AbstractProcess,
    Dict,
    ImageProcessingPipeline,
    IncompatibleArgsException,
    ProcessFork,
    ProcessJoined,
    ProcessLogic,
    ProcessLogicProperty,
    ProcessPassThrough,
)

# filepath: pyMOF/image_processing/test_pipelines.py


def test_dict_missing_key():
    d = Dict()
    assert d['missing_key'] is None
    assert d.missing_key is None


class TestIncompatibleProcess1(AbstractProcess):
    def __call__(self, image, **kwargs) -> Dict:
        return Dict(image=image, result="test1")


class TestIncompatibleProcess2(AbstractProcess):
    def __call__(self, contour, **kwargs) -> Dict:
        return Dict(contour=contour, result="test2")


def test_incompatible_args_exception():
    with pytest.raises(IncompatibleArgsException) as excinfo:
        process1 = TestIncompatibleProcess1()
        process2 = TestIncompatibleProcess2()
        joined = process1 >> process2
        _ = joined(image=None)
    assert (str(excinfo.value) ==
            "The process 'TestIncompatibleProcess2' received incompatible"
            " payload from the previous process 'TestIncompatibleProcess1'.")


class TestProcess(AbstractProcess):
    def __call__(self, **kwargs) -> Dict:
        return Dict(result="test")


class TestAlternativeProcess(AbstractProcess):
    def __call__(self, **kwargs) -> Dict:
        return Dict(result="test")


def test_process_logic():
    @processLogic
    def logic_func(value, **kwargs):
        return Dict(value=value, result="logic")

    process = logic_func()
    assert isinstance(process, ProcessLogic)
    result = process(value="value")
    assert result == Dict(value="value", result="logic")


def test_process_logic_property():
    class AClass:
        @processLogicProperty
        def logic_func(self, value, **kwargs):
            return Dict(value=value, result="logic")

    c = AClass()
    process = c.logic_func
    assert isinstance(process, ProcessLogicProperty)
    result = process(value="value")
    assert result == Dict(value="value", result="logic")


def test_process_factory():
    @processFactory(cache=False)
    def factory_func(flg: bool):
        if flg:
            return TestProcess()
        else:
            return TestAlternativeProcess()

    process = factory_func(flg=True)
    assert isinstance(process, TestProcess)
    process2 = factory_func(flg=False)
    assert isinstance(process2, TestAlternativeProcess)


def test_cached_process_factory():
    @processFactory(cache=True)
    def factory_func(flg: bool):
        if flg:
            return TestProcess()
        else:
            return TestAlternativeProcess()

    process = factory_func(flg=True)
    assert isinstance(process, TestProcess)
    process2 = factory_func(flg=False)
    assert isinstance(process2, TestAlternativeProcess)
    # call again
    process3 = factory_func(flg=True)
    assert isinstance(process3, TestProcess)
    process4 = factory_func(flg=False)
    assert isinstance(process4, TestAlternativeProcess)


def test_process_pass_through():
    process = ProcessPassThrough()
    result = process(test="value")
    assert result == Dict(test="value")


def test_process_joined():
    process1 = TestProcess()
    process2 = TestProcess()
    fork = ProcessFork([process1, process2])
    joined = ProcessJoined(fork)
    result = joined()
    assert result == Dict(result="test")


def test_process_fork():
    process1 = TestProcess()
    process2 = TestProcess()
    fork = ProcessFork([process1, process2])
    result = fork()
    assert result == (Dict(result="test"), Dict(result="test"))


def test_image_processing_pipeline():
    process1 = TestProcess()
    process2 = TestProcess()
    pipeline = ImageProcessingPipeline([process1, process2])
    result = pipeline()
    assert result == Dict(result="test")


def test_process_rshift():
    process1 = TestProcess()
    process2 = TestProcess()
    pipeline = process1 >> process2
    assert isinstance(pipeline, ImageProcessingPipeline)
    result = pipeline()
    assert result == Dict(result="test")


def test_process_mul():
    process1 = TestProcess()
    process2 = TestProcess()
    fork = process1 * process2
    assert isinstance(fork, ProcessFork)
    result = fork()
    assert result == (Dict(result="test"), Dict(result="test"))


def test_process_fork_rshift():
    process1 = TestProcess()
    process2 = TestProcess()
    process3 = TestProcess()
    fork = process1 * process2
    pipline = fork >> process3
    assert isinstance(pipline, ImageProcessingPipeline)
    assert isinstance(pipline.processes[0], ProcessJoined)
    result = pipline()
    assert result == Dict(result="test")


def test_process_join_fork_by_empty_dict():
    process1 = TestProcess()
    process2 = TestProcess()
    fork = process1 * process2
    joined = fork / {}
    assert isinstance(joined, ProcessJoined)
    result = joined()
    assert result == Dict(result="test")


def test_process_fork_truediv():
    process1 = TestProcess()
    process2 = TestProcess()
    fork = process1 * process2
    joined = fork / {0: [("result", "result1")], 1: [("result", "result2")]}
    assert isinstance(joined, ProcessJoined)
    result = joined()
    assert result == Dict(result1="test", result2="test")


def test_image_processing_pipeline_rshift():
    process1 = TestProcess()
    process2 = TestProcess()
    pipeline1 = ImageProcessingPipeline([process1])
    pipeline2 = pipeline1 >> process2
    assert isinstance(pipeline2, ImageProcessingPipeline)
    result = pipeline2()
    assert result == Dict(result="test")


def test_image_processing_pipeline_mul():
    process1 = TestProcess()
    process2 = TestProcess()
    pipeline = ImageProcessingPipeline([process1])
    fork = pipeline * process2
    assert isinstance(fork, ProcessFork)
    result = fork()
    assert result == (Dict(result="test"), Dict(result="test"))



"""The core module.

   Notes:
   ------

    1- Service handler's arguments
        Service's 'handle_event' arguments can be provided by
        the caller on binding time. To do that, you must set the
        Event's extra_kwargs : mapping[str, ServiceCallable | Any ].

        Mapping's key is the name of the argument. Mapping's value can
        be passed as Any, or can be a ServiceCallable type to be loaded
        lazily: 'type  ServiceCallable = Callable[[None | object], Any]'

        In other words, in the former case, the same value will send
        to to service every time that the 'handle_event' is called,
        and in the latter case, the Callable updates the value of
        the argument each time the 'handle_event' is called.
        The ServiceCallable signutare MUST have zero argument, or for class
        function, the 'self' argument.

        It can raise 'ServiceArgumentError'.

    2- Service handler's arguments as callback
        When the service argument is a callback, which will be called
        later by the 'handle_event', it MUST be decorated by @service_callback.
        It becomes clear, if we  consider that the arguments can be lazily
        loaded, and so, it can be a Callable. As a result, on calling the
        service's 'handle_event', it must be differentiated between Callable
        as lazy load or callback.

        It can raise 'ServiceArgumentError'.

    3- Configurable
       Both AbstractApp , AbstractLayout and TogaComponent
       are 'Configurable'. So, they can impliment
       os dependent configuration.

       AbstractApp calls 'Configurable' once in its
       lifcycle:  on_begin().

       AbstractLayout calls 'Configurable' once on its
       first load: on_load.
"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import Enum
from inspect import signature
from typing import Any, Callable, Mapping

from pyMOSF.config import Configurable, GUIFramework, Settings
from pyMOSF.core.__loggers__ import file_out_log, std_out_log

log = logging.getLogger(__name__)


class ServiceArgumentError(Exception):
    def __init__(self, message: str, cause=None) -> None:
        self.message = message
        self.__cause__ = cause
        super().__init__(self.message)


class EventType(Enum):
    ON_PRESS = 0
    ON_RELEASE = 1
    ON_CHANGE = 2
    ON_TOUCH_DOWN = 3
    ON_TOUCH_MOVE = 4
    ON_TOUCH_UP = 5
    BIND = 100

    @staticmethod
    def get_all_eventTypes():
        return vars(EventType)['_member_names_']


class Updateable:
    def on_update(self, **kwargs):
        """Update the component with new data."""
        pass


class Service(ABC):
    """An abstract class for handeling UI event callbacks
    """

    @abstractmethod
    def handle_event(self,
                     widget: Any,
                     app,
                     service_callback: Callable | None = None,
                     *args,
                     **kwargs) -> None:
        """An abstract class for handeling UI sync event callbacks
        """
        pass

    def on_exit(self) -> None:
        """The event handler that will be invoked when the app is about to exit.
        """
        pass


# Accept zero or one argument
# type  ServiceCallback = Callable[[object | Dict], None]


class Event:
    def __init__(self,
                 id: str,
                 eventType: EventType,
                 service: Service,
                 property_name: str = "",
                 extra_kwargs: Mapping[str, Any] = {},
                 service_callback=None
                 ) -> None:
        self.id = id
        self.eventType = eventType
        self.service = service
        self.property_name = property_name
        self.extra_kwargs = extra_kwargs
        self.service_callback = service_callback

    def element_event(self, element):
        return getattr(element, self.eventType.name.lower())


class AbstractLayout(ABC, Configurable,  Updateable):
    def __init__(self, app: AbstractApp) -> None:
        super(AbstractLayout, self).__init__()
        self._app = app

    @property
    def _name(self):
        return type(self).__qualname__

    @abstractmethod
    def build_layout(self, app) -> Any:
        pass

    @property
    def ml_app(self) -> AbstractApp:
        if self._app is None:
            raise ValueError("app has not been initialized.")
        return self._app

    def on_load(self):
        pass

    def on_end(self):
        pass


class AbstractApp(ABC, Configurable):
    def __init__(self,
                 layout: AbstractLayout,
                 settings: Settings | None = None,
                 **kwargs) -> None:
        super(AbstractApp, self).__init__(**kwargs)
        self._settings = settings
        self.layout = layout
        self.reset_event_dispatchers_table()

    def reset_event_dispatchers_table(self):
        self._event_dispatchers_table = {}
        for t in EventType.get_all_eventTypes():
            self._event_dispatchers_table[t] = {}

    @property
    def settings(self) -> Settings:
        if self._settings is None:
            self._settings = Settings.load(self.data_path)
        return self._settings

    @property
    def event_dispatchers_table(self):
        """Stores different event_dispatchers per EventType.

        event_dispatchers_table is dictionary or dictionaries that
        EventType is its key and event_dispatchers are dictionaries
        of (id : UI element).
        """
        return self._event_dispatchers_table

    @property
    def path(self):
        raise NotImplementedError()

    @property
    def data_path(self):
        raise NotImplementedError()

    def on_begin(self):
        """The event handler that will be invoked when the app is about to start.

        It calls only once, in contrast to 'on_load_window', therefore, app settings
        is called once during the life-cycle of an app.

        It calls the build_layout of Layout instance, stores all UI elements with
        id and suport of one or more events in event_dispatchers_table, and registers
        them in ServiceRegistry.
        """
        self._set_config()
        self.on_load()

    def on_load(self):
        """The event handler that will be invoked when the app is reset the layout.

        It can be called any time the layout or main window chances, in contrast to
        'on_begin'.

        It calls the build_layout of Layout instance, stores all UI elements with
        id and suport of one or more events in event_dispatchers_table, and registers
        them in ServiceRegistry.
        """
        root = self.layout.build_layout(self)

        # self.reset_event_dispatchers_table()
        self.__enumerate_elements(root)
        # Initial Event bindings definitions happens here
        registry = ServiceRegistry()
        #
        previous_events = [id for id in registry.events]
        self.layout._set_config()
        # Event bindings to dispatcher happens here

        for id, event in registry.events.items():
            event_dispatcher = self.event_dispatchers_table[event.eventType.name]
            if id in previous_events:
                continue
            if id not in event_dispatcher:
                raise ValueError(
                    f" Widget id:'{id}' for the {event.eventType.name} handler "
                    f"has no corresponding UI. Check the bindings map.")
            element = event_dispatcher[id]
            registry.register_service(event,
                                      element,
                                      app=self)
        #
        self.layout.on_load()

    def on_end(self) -> bool:
        """The event handler that will be invoked when the app is about to exit.

        It calls the on_exit of all services in the ServiceRegistry.
        """
        registry = ServiceRegistry()
        registry.on_exit(self)
        self.layout.on_end()
        self.settings.on_end()
        return True

    def __enumerate_elements(self, root):
        """Recursivly searches and records all elements.
        """

        def creat_method_checker(name):
            """Create a function that checks availibility of a method
            """
            def checker(obj):
                """A combinations of attribute name and callable"""
                method = getattr(obj, name, None)
                return method is not None and callable(method)

            return checker

        event_checkers = [(t, creat_method_checker(t.lower()))
                          for t in EventType.get_all_eventTypes()]

        def recursive(parent):
            """Recusricly check all UI's tree elements.

               If the element has both an id and a corresponding
               event, it will be stored in event_dispatchers.
            """
            for child in parent.children:
                # event_checkers contains both the eventType and corresponding
                # method (has_event) that check the availibity of the event.
                # e.g. for EventType.ON_PRESS, has_event looks for UI element's
                # on_press event support.
                for eventType, has_event in event_checkers:
                    if has_event(child) and hasattr(child, "id") is True:
                        # event_dispatchers_table is dictionary or dictionaries
                        event_dispatchers = self.event_dispatchers_table[eventType]
                        event_dispatchers[child.id] = child
                recursive(child)

        recursive(root)

    def update(self, **kwargs):
        """Update the component with new data."""
        self.layout.on_update(**kwargs)

    def set_logger(self,
                   std_out: bool = True,
                   file_out: bool = False,
                   file_name: str = "log.txt",
                   log_level: int = logging.ERROR):
        if std_out:
            std_out_log(log_level=logging.getLevelName(log_level))
        if file_out:
            file_out_log(path=self.data_path / file_name,
                         log_level=logging.getLevelName(log_level))


class ServiceRegistry(object):
    """A singleton object that store the required (id: ServiceStrategy).
    """
    _instance = None
    _framework: GUIFramework = GUIFramework.UKNOWN

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ServiceRegistry, cls).__new__(cls)
            # Put any initialization here.
            cls._instance._dispatcher = EventDispatcher()
        return cls._instance

    __events = {}
    _dispatcher: EventDispatcher

    @property
    def dispatcher(self):
        return self._dispatcher

    @property
    def events(self):
        return self.__events

    def bind(self,
             *,
             id: str,
             eventType: EventType,
             service,
             property_name: str = "",  # kivy only, to be used by its bind
             extra_kwargs: Mapping[str, Any] = {},
             service_callback=None) -> None:
        self.events[id] = Event(id, eventType, service,
                                property_name, extra_kwargs,
                                service_callback)

    def bind_event(self, event: Event) -> None:
        self.events[event.id] = event

    def get_event_info(self, id) -> Event:
        return self.events.get(id, None)  # type: ignore

    def register_service(self, event: Event, element: Any, app: AbstractApp):
        """Register a ServiceStrategy for the given event, UI element and app.

           A callback will be attached to the event handler of the UI element.
           If the UI element has also a handler, it will be included in the call
           qeue.


        Parameters
        ----------
        event : Event
            The event object that includes id and ServiceStrategy.
        element : Any
            UI element that can handel th event.
        app : AbstractApp
            The app object.
        """
        if event.extra_kwargs is not {} and not isinstance(event.extra_kwargs, Mapping):
            raise ValueError(
                f" Widget id:'{event.id}' ({type(element).__name__}) "
                f"asked for {event.eventType.name} handler, but did not provide "
                f"the binding arguments (Mapping). Check the bindings map.")
        #

        def attach(callback):
            # setattr(element, registeredEventType.name.lower(), callback)
            match ServiceRegistry._framework:
                case GUIFramework.TOGA:
                    # from toga.handlers import wrapped_handler  # type: ignore

                    handler = callback  # wrapped_handler(element, callback)
                    # handler = wrapped_handler(element, callback)
                    # e.g. element.on_press = handler
                    setattr(element, event.eventType.name.lower(), handler)
                case GUIFramework.KIVY:
                    # Only BIND
                    if event.eventType == EventType.BIND:
                        if event.property_name == "":
                            raise ValueError(
                                f" Widget id:'{event.id}' ({type(element).__name__}) "
                                f"asked for {event.eventType.name} handler, but did not "
                                f"provide the 'property_name'. Check the bindings map.")
                        if not isinstance(event.property_name, str):
                            raise ValueError(
                                f" Widget id:'{event.id}' ({type(element).__name__}) "
                                f"asked for {event.eventType.name} handler, but did not provide "
                                f"the binding argument name (str). Check the bindings map.")
                        # e.g. switch.bind(active=callback)
                        kwrgs = {event.property_name: callback}
                        element.bind(**kwrgs)
                    else:
                        # e.g. button.bind(on_press=callback)
                        kwrgs = {event.eventType.name.lower(): callback}
                        element.bind(**kwrgs)
                case _:
                    raise ValueError(
                        f" Unknown UI Framework: {ServiceRegistry._framework}")
        #
        element_event = event.element_event(element)
        if hasattr(element_event, "_raw"):
            handler = element_event._raw  # This is toga's wrapped_handler attribute
        else:
            handler = element_event  # kivy

        #
        match event.service:
            case SyncService():
                if handler is not None:
                    self.dispatcher.register_framework(event.id, handler)
                self.dispatcher.register(event.id, event.service.handle_event)
                handler_2 = self.dispatcher.service_callback(event.id,
                                                             element,
                                                             app,
                                                             event.service_callback,
                                                             event.extra_kwargs)
                attach(handler_2)
            case AsyncService():
                if handler is not None:
                    self.dispatcher.register_async_framework(event.id, handler)
                self.dispatcher.register_async(
                    event.id, event.service.handle_event)
                handler_2 = self.dispatcher.service_async_callback(event.id,
                                                                   element,
                                                                   app,
                                                                   event.service_callback,
                                                                   event.extra_kwargs)
                attach(handler_2)
            case _:
                log.warning(
                    f" Widget id:'{event.id}' ({type(element).__name__}) "
                    f"did not bind to any {event.eventType.name} handler. Check the bindings map.")

    def on_exit(self, app):
        """The event handler that will be invoked when the app is about to exit.

        It calls the on_exit of all registred services.
        """
        for eventInfo in self.events.values():
            eventInfo.service.on_exit()

        # for id, event in self.events.items():
        #     event_dispatcher = app.event_dispatchers_table[event.eventType.name]
        #     element = event_dispatcher[id]
        #     print(element)

    def fire_event(self, id, app):
        event = self.get_event_info(id)
        event_dispatcher = app.event_dispatchers_table[event.eventType.name]
        element = event_dispatcher[id]
        callback = self.dispatcher.service_callback(event.id,
                                                    element,
                                                    app,
                                                    event.service_callback,
                                                    event.extra_kwargs)
        callback()

    async def fire_async_event(self, id, app):
        event = self.get_event_info(id)
        event_dispatcher = app.event_dispatchers_table[event.eventType.name]
        element = event_dispatcher[id]
        callback = self.dispatcher.service_async_callback(event.id,
                                                          element,
                                                          app,
                                                          event.service_callback,
                                                          event.extra_kwargs)
        await callback()


class EventDispatcher:
    """Event dispatcher that binds the UI elements call back to EventStrategy instances.

    This is a lightweight event dispatcher that can support both sync and async calls.
    """

    def __init__(self):
        self.reset_listeners()

    def reset_listeners(self):
        self.listeners = {}
        self.listeners_framework = {}
        self.async_listeners = {}
        self.async_listeners_framework = {}

    #
    def register(self, event_name: str, handler):
        """Register a synchronous callback for an event.

        Parameters
        ----------
        event_name : str
            The name of the event to register.
        handler : function
            The function to call when the event is triggered.
        """
        if event_name not in self.listeners:
            self.listeners[event_name] = []
        self.listeners[event_name].append(handler)

    def register_framework(self, event_name: str, handler):
        """Register a synchronous callback for an event that has already binded in the framework.

        Parameters
        ----------
        event_name : str
            The name of the event to register.
        handler : function
            The function to call when the event is triggered.
        """
        if event_name not in self.listeners_framework:
            self.listeners_framework[event_name] = []
        self.listeners_framework[event_name].append((handler,
                                                     signature(handler)))

    def register_async(self, event_name: str, async_handler):
        """Register an asynchronous callback for an event.

        Parameters
        ----------
        event_name : str
            The name of the event to register.
        async_handler : function
            The async function to call when the event is triggered.
        """

        if event_name not in self.async_listeners:
            self.async_listeners[event_name] = []
        self.async_listeners[event_name].append(async_handler)

    def register_async_framework(self, event_name: str, async_handler):
        """Register an asynchronous callback for an event that has already binded in the framework.

        Parameters
        ----------
        event_name : str
            The name of the event to register.
        async_handler : function
            The async function to call when the event is triggered.
        """

        if event_name not in self.async_listeners_framework:
            self.async_listeners_framework[event_name] = []
        self.async_listeners_framework[event_name].append((async_handler,
                                                           signature(async_handler)))
    #

    def dispatch(self,
                 event_name: str,
                 widget: Any,
                 app: AbstractApp,
                 call_back,
                 extrakwargs: Mapping,
                 *args,
                 **kwargs):
        """Dispatch a synchronous event to all registered callbacks.

        Parameters
        ----------
        event_name : str
            The name of the event to dispatch.
        widget : Any
            UI element.
        app : AbstractApp
            The AbstractApp subclass that registred the elements.
        call_back: ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        extrakwargs : Mapping
            Extra key-value paires can be provided to pass to the handler.
            If the value is a callable, it is called each time the event happens.
            This is useful when the arguments need to be updated, e.g., from the
            value in the UI.
        """
        if ServiceRegistry._framework == GUIFramework.TOGA:
            for binded_method, sig in self.listeners_framework.get(event_name, []):
                binded_method(*args, **kwargs)

        for binded_method in self.listeners.get(event_name, []):
            if extrakwargs is None:
                binded_method(widget, app, call_back, *args, **kwargs)
            else:
                kwargs2 = {}
                for k, v in extrakwargs.items():
                    if callable(v):
                        parameters = signature(v).parameters
                        if len(parameters) > 1:
                            raise ServiceArgumentError(
                                message=f"In setting the parameter '{k}', the signuture of "
                                f"function '{v.__name__}' for service handler "
                                f"of the {event_name=} must have zero or self argument, it needs "
                                f"{len(parameters):d} = {list(parameters.keys())}. "
                                "(You might need to decorate it by @component_method."
                            )
                        else:
                            try:
                                kwargs2[k] = v()
                            except TypeError as e:
                                raise ServiceArgumentError(
                                    message="Service lazy argument caused an error. "
                                    "check the function accept zero or at most one "
                                    "argument, 'self'.",
                                    cause=e
                                )

                    else:
                        kwargs2[k] = v

                binded_method(widget, app, call_back, **kwargs2)

    async def dispatch_async(self,
                             event_name: str,
                             widget: Any,
                             app: AbstractApp,
                             call_back,
                             extrakwargs: Mapping,
                             *args,
                             **kwargs):
        """Dispatch a asynchronous event to all registered callbacks.

        Parameters
        ----------
        event_name : str
            The name of the event to dispatch.
        widget : Any
            UI element.
        app : AbstractApp
            The AbstractApp subclass that registred the elements.
        call_back: ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        extrakwargs : Mapping
            Extra key-value paires can be provided to pass to the handler.
            If the value is a callable, it is called each time the event happens.
            This is useful when the arguments need to be updated, e.g., from the
            value in the UI.
        """
        if ServiceRegistry._framework == GUIFramework.TOGA:
            for binded_method, sig in self.async_listeners_framework.get(event_name, []):
                binded_method(*args, **kwargs)
        for binded_method in self.async_listeners.get(event_name, []):
            if extrakwargs is None:
                await binded_method(widget, app, call_back, *args, **kwargs)
            else:
                kwargs2 = {}
                for k, v in extrakwargs.items():
                    if callable(v):
                        parameters = signature(v).parameters
                        if len(parameters) > 1:
                            raise ValueError(
                                f"In setting the parameter '{k}', the signuture of function "
                                f"'{v.__name__}' for service handler "
                                f"of the {event_name=} must have zero or self argument, it needs "
                                f"{len(parameters):d} = {list(parameters.keys())}. "
                                "(You might need to decorate it by @component_method "
                                "or @service_callback)."
                            )
                        else:
                            try:
                                kwargs2[k] = v()
                            except TypeError as e:
                                raise ServiceArgumentError(
                                    message="Service lazy argument caused an error. "
                                    "check the function accept zero or at most one "
                                    "argument, 'self'.",
                                    cause=e
                                )

                    else:
                        kwargs2[k] = v

                await binded_method(widget, app, call_back, **kwargs2)

    def service_callback(self,
                         id: str,
                         widget: Any,
                         app: AbstractApp,
                         call_back,
                         extrakwargs: Mapping):
        """Create a synchronous widget callback for a ServiceStrategy instance.

        Parameters
        ----------
        id : str
            The ID of the widget.
        widget : Any
            The UI element that registers the and will call the callback.
        app : AbstractApp
            The application instance.
        call_back: ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        extrakwargs : Mapping
            Extra argument that maybe provided on binding table.

        Returns
        -------
        function
            A decorated callback function.
        """
        def callback(*args, **kwargs):
            self.dispatch(id, widget, app, call_back,
                          extrakwargs, *args, **kwargs)

        return callback

    def service_async_callback(self,
                               id: str,
                               widget: Any,
                               app: AbstractApp,
                               call_back,
                               extrakwargs: Mapping):
        """Create an asynchronous widget callback for a ServiceStrategy instance.

        Parameters
        ----------
        id : str
            The ID of the widget.
        widget : Any
            The UI element that registers the and will call the callback.
        app : AbstractApp
            The application instance.
        call_back: ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        extrakwargs : Mapping
            Extra argument that maybe provided on binding table.

        Returns
        -------
        function
            A decorated async callback function.
        """
        async def async_callback(*args, **kwargs):
            await self.dispatch_async(id, widget, app, call_back, extrakwargs, *args, **kwargs)

        return async_callback


class SyncService(Service):
    """An abstract class for handeling UI sync event callbacks
    """

    @abstractmethod
    def handle_event(self,
                     widget: Any,
                     app: AbstractApp,
                     service_callback,
                     *args,
                     **kwargs) -> None:
        """Sync event handler

        Parameters
        ----------
        widget : Any
            The widget that called the handler.
        app : AbstractApp
            The parent application of the widget.
        service_callback : ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        """
        pass


class AsyncService(Service):
    """An abstract class for handeling UI async event callbacks
    """

    @abstractmethod
    async def handle_event(self,
                           widget: Any,
                           app: AbstractApp,
                           service_callback,
                           *args,
                           **kwargs) -> None:
        """Async event handler.

        Parameters
        ----------
        widget : Any
            The widget that called the handler.
        app : AbstractApp
            The parent application of the widget.
        service_callback : ServiceCallback | None
            The callback function that will be called after the event to
            send the possible result.
        """
        pass



"""
Studying bees diseases by analysing photos
"""
import logging

from pyMOSF.templates.toga.example_layouts import ExampleViewLayout
from pyMOSF.toga import TogaMultiLayoutApp

log = logging.getLogger(__name__)


class TogaApp(TogaMultiLayoutApp):
    """A toga App class that has a layout.

    The layout class will create UI elements and its main_box will be added
    as the app main window's content.
    """

    def __init__(self,
                 formal_name: str,
                 app_id="net.pazuki.torchApp"):

        self.layout = ExampleViewLayout(self)

        super(TogaApp, self).__init__(init_layout=self.layout,
                                      formal_name=formal_name,
                                      app_id=app_id,
                                      )

    def startup(self):
        super().startup()
        self.show_layout(self.layout)


def create_app():
    log.info("Eample app stars")

    app = TogaApp("Example app")
    return app


if __name__ == "__main__":
    # This main entry point is used for debugging.
    # It is not used in the main app.
    log.info("Debug main is called.")
    try:
        create_app().main_loop()
    except Exception as e:
        print(e)
        log.error("Exception in main loop. \n"
                  f"Exception: {e}"
                  f"\nTraceback: {e.__traceback__}")



from __future__ import annotations

import ast
import inspect
from abc import abstractmethod
from typing import TypeVar

from pyMOSF.config import Settings
from pyMOSF.core.__core__ import AbstractApp, AbstractLayout, Configurable, Updateable

Box = TypeVar("Box")


class Layout(AbstractLayout):

    def __init__(self, app: MultiLayoutApp) -> None:
        self._app = app
        super().__init__(app)

    @property
    def ml_app(self) -> MultiLayoutApp:
        return self._app

    @property
    def main_box(self):
        if self._box is None:
            raise ValueError("box has not been initialized (_build_box).")
        return self._box

    def build_layout(self, app):
        """Build the main window of the app and its layout.
        """
        #
        self._app = app
        self._box = self._build_box()
        #
        return self._box

    @abstractmethod
    def _build_box(self):
        """Build the layout box and all the required UI elements.

           Also, all the buttons that has an id registred in ServiceRegistry
           will be bind to on_press handlers.

        Parameters
        ----------
        image_view : toga.ImageView
            Every app have and ImageView that the layout class must
            include it in a box.
        """
        pass


class Component(Configurable, Updateable):
    def __init__(self, layout: StackedLayout, **kwargs) -> None:
        self._layout = layout
        super(Component, self).__init__(**kwargs)

    @property
    def ml_app(self) -> MultiLayoutApp:
        return self._layout.ml_app

    @property
    def parent_layout(self) -> StackedLayout:
        return self._layout

    def on_load(self):
        pass

    def on_end(self):
        pass


T = TypeVar("T", bound=Component)
Ts = TypeVar("Ts", bound=Component)


def _get_component_passthrough(cls):
    target = cls
    component_methods = []

    def visit_FunctionDef(node):
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(
                    n.func, ast.Attribute) else n.func.id  # type: ignore
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id
            if name == 'component_method':
                component_methods.append(node.name)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(inspect.getsource(target)))

    return component_methods


class StackedLayout(Layout):

    def __init__(self,
                 app: MultiLayoutApp,
                 box,
                 *types):  # type: ignore
        assert len(types) > 0, "At least one KivyComponent type is needed."
        self._box = box
        self.types = types
        self._instances = {}
        super(StackedLayout, self).__init__(app)

    def __getitem__(self, key: type[T]) -> T:
        return self.instances[key.__name__]

    def _init_component(self, component_type: type[T]) -> T:
        """override this method if there are component specific initilisations.

        Parameters
        ----------
        component_type : type[T]
            Component's type

        Returns
        -------
        T
            An instance of the component.
        """
        return component_type(layout=self)

    @property
    def instances(self) -> dict:
        if self._instances == {}:
            def name(component_type) -> str: return component_type.__name__
            self._instances = {
                name(component_type): self._init_component(component_type)
                for component_type in self.types
            }

        return self._instances

    @property
    def _name(self) -> str:
        return ":".join(t.__qualname__ for t in self.types)

    def build_layout(self, app):
        #
        self._app = app
        self._build_box()
        #
        return self._box

    @abstractmethod
    def _add_to_main_box(self, box: Box, element: Box):
        pass

    def _build_box(self):
        """Construct and show the content layout of Toga application."""
        #
        for instance, component_type in zip(self.instances.values(), self.types):
            self._add_to_main_box(self._box, instance)
            # passthrough the methods of the components to the layout
            # find the components methods that has @component_method decorator
            methods = _get_component_passthrough(component_type)
            for name in methods:
                # define the same method for the layout
                m = getattr(instance, name)
                setattr(self, name, m)
            #
            instance.on_load()
        return self._box

    def _set_config(self):
        """Layout related config (e.g. bindings).
        """
        #
        super()._set_config()
        for instance in self.instances.values():
            instance._set_config()

    def on_end(self):
        for component in self.instances.values():
            component.on_end()

    def on_update(self, **kwargs):
        """Update the component with new data."""
        for instance in self.instances.values():
            instance.on_update(**kwargs)
        super().on_update(**kwargs)


class MultiLayoutApp(AbstractApp):

    def __init__(self,
                 main_container,
                 init_layout: Layout,
                 settings: Settings | None = None,
                 **kwargs) -> None:
        if not isinstance(init_layout, Layout):
            raise ValueError(
                f"The {init_layout=} argument must be a subclass of 'TogaLayout'.")
        #
        self._main_container = main_container
        self._layout_lookup = {init_layout._name: init_layout}
        self._current_layout = init_layout
        #
        super(MultiLayoutApp, self).__init__(
            init_layout, settings, **kwargs)

    @property
    def main_container(self):
        return self._main_container

    @property
    def current_layout(self):
        return self._current_layout

    @abstractmethod
    def _add_layout(self, layout):
        pass

    def show_layout(self, layout):
        if layout._name not in self._layout_lookup:
            self._layout_lookup[layout._name] = layout
            self.layout = layout
            #
            # Call on_load in AbstractApp to register the event and
            # build to layout. It must be called once
            # Layout's on_load() will be called inside this function
            self.on_load()
            #
            self._add_layout(self.layout)
        else:
            layout = self._layout_lookup[layout._name]
            #
            self._add_layout(layout)
            #
            self.layout = layout
            # Call layout's on_load for any reloading logic
            self.layout.on_load()
            #

    def on_end(self) -> bool:
        for layout in self._layout_lookup.values():
            layout.on_end()
        return super().on_end()

    def update(self, **kwargs):
        """Update the component with new data."""
        self.current_layout.on_update(**kwargs)



import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Mapping, TypeVar

from pyMOSF.config import Dict

_Self = TypeVar('_Self', bound='AbstractPipeline')


# logger
log = logging.getLogger(__name__)


class IncompatibleArgsException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class AbstractPipeline(ABC):
    """An abstract pipeline class.
    """

    def __init__(self, processes=[]):
        self.processes = processes

    def append_process(self, process) -> None:
        self.processes.append(process)

    def __call__(self, /, **kwargs) -> Dict:
        return self.process(**kwargs)

    @abstractmethod
    def process(self, /, **kwargs) -> Dict:
        pass

    @abstractmethod
    def __rshift__(self: _Self, other) -> _Self:
        pass

    @abstractmethod
    def __mul__(self, other) -> 'ProcessFork':
        pass


class AbstractProcess(ABC):

    @abstractmethod
    def __call__(self, **kwargs) -> Dict:
        """The operation of the process must happen here.

        It can define any number of arguments it like and
        must have one '**kwargs' at the end.

        The returns will be named payload as a Dict object
        and will be passed to the down-stream process or
        return to the caller.
        """
        pass

    def __rshift__(self, other) -> AbstractPipeline:
        """Appends the process to the end of an Process or ImageProcessingPipeline.

           This is an imumutable operation and new ImageProcessingPipeline will be
           returned.

           Example:
                proc1 >> proc2 >> proc3

        Parameters
        ----------
        other : AbstractProcess | ImageProcessingPipeline
            The process or pipeline that will be appended to the given
            process.
            For ProcessFork in the RHS, the output of the ProcessFork
            will be joined by returning a ProcessJoined object.

        Returns
        -------
        AbstractPipeline
            A new (immutable) pipeline object that contains the current
            process followed by 'other' argument.

        Raises
        ------
        ValueError
            Arises when the RHS is not one of the accptable types.
        """
        if issubclass(type(other), AbstractProcess):
            if isinstance(other, ProcessFork):
                other = ProcessJoined(other)
            return ImageProcessingPipeline([self, other])
        elif issubclass(type(other), AbstractPipeline):
            return ImageProcessingPipeline(
                [self] + other.processes)
        else:
            raise ValueError(
                f"The '{type(other)}' must be a AbstractProcess or ImageProcessingPipeline.")

    def __mul__(self, other):
        """Fork two or more processes.

           This is an imumutable operation and new ProcessFork will be
           returned.

           Example:
               p1 * p2 : Forks two processes
               p1 * p2 * p3 : Forks three processes

        Parameters
        ----------
        other : int, ProcessFork, AbstractProcess, ImageProcessingPipeline
            When the RHS is n (int), it is forked to n replicate.
            When the RHS is ProcessFork, AbstractProcess or ImageProcessingPipeline,
            the LHS process stack on its process.

        Returns
        -------
        ProcessFork
            A callable object that forks the processes.

        Raises
        ------
        ValueError
            Raises when the 'other' argument is not an int, a ProcessFork, sub class of
            AbstractProcess, or ImageProcessingPipeline.
            Also, in case for RHS int and LHS ProcessFork, it will raise the error.
        """
        if isinstance(other, int):
            if isinstance(self, ProcessFork):
                raise ValueError(
                    "The 'ProcessFork' has already forked (cannot be multiplied).")
            else:
                return ProcessFork([self] * other)
        if isinstance(other, ProcessFork):
            if isinstance(self, ProcessFork):
                return ProcessFork(self.processes + other.processes)
            else:
                return ProcessFork([self] + other.processes)
        if issubclass(type(other), AbstractProcess):
            return ProcessFork([self, other])
        if issubclass(type(other), ImageProcessingPipeline):
            return ProcessFork([self, other])
        else:
            raise ValueError(
                f"The '{type(other)}' must be an int, a AbstractProcess, ImageProcessingPipeline "
                f"or tuples of both.")


class Process(AbstractProcess):
    """A singleton object for instantiable processes.
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Process, cls).__new__(cls)
            # Put any initialization here.
        return cls._instance


class ProcessPassThrough(Process):
    def __call__(self, **kwargs) -> Dict:
        """Payload is simplely pass to the next process."""
        return Dict(**kwargs)


class ProcessJoined(AbstractProcess):
    def __init__(self,
                 forkedProcess: AbstractProcess,
                 kwargs_mapping: Mapping[int, list[tuple[str, str]]] = {}):
        """Join all the forked processes and returns as a single Mapping.

        Parameters
        ----------
        forkedProcess : Process
            The process that has been forked.
        kwargs_mapping : Mapping[int, list[tuple[str, str]]], optional
            Provides the renaming of the processes outputs, by default None.
            The key is the index of the process in the forked process.
            The value is a list of tuples of the old key and the new key.
        """
        self.forkedProcess = forkedProcess
        self.kwargs_mapping = kwargs_mapping

    def __call__(self, **kwargs) -> Dict:
        # First, calls all the sub-processes in the fork.
        tuple_return = self.forkedProcess(**kwargs)
        # Next, if there is any renaming requested in the
        # payloads of the forked sub-processes, we do it here
        for index, names in self.kwargs_mapping.items():
            ret = tuple_return[index]
            for old_key, new_key in names:
                ret[new_key] = ret.pop(old_key)
        # Finally, make a union of all the retuned payloads.
        # Note: the names from returns of higher rank  has precedence
        new_kwargs = tuple_return[0]
        for d in tuple_return[1:]:
            new_kwargs |= d
        return new_kwargs


class ProcessFork(AbstractProcess):
    def __init__(self,
                 processes: list):
        assert len(processes) > 1, "The processes must be more than one."
        self.processes = processes

    def __getitem__(self, key: int) -> AbstractProcess | AbstractPipeline:
        return self.processes[key]

    def __setitem__(self, key: int, process) -> None:
        self.processes[key] = process

    def __call__(self,  **kwargs) -> tuple[Dict, ...]:
        """It calles each sub-processes of the fork and returns thier payload as a tuple."""
        return tuple(p(**kwargs) for p in self.processes)

    def __rshift__(self, other) -> ProcessJoined:
        # It first join the forked instance, and next,
        # calls the abstractProcess rshift to append the
        # RHS to the joined process.
        return ProcessJoined(self) >> other

    def __itruediv__(self, other) -> ProcessJoined:
        return self.__truediv__(other)

    def __truediv__(self, other: Mapping[int, list[tuple[str, str]]]) -> ProcessJoined:
        """combine the output of a forked process.

            Examples:
            forked_process /= {0:[ ('old_name1', 'new_name1'),
                                   ('old_name2', 'new_name2')],
                               1:[ ('old_name3', 'new_name3'),
                                   ('old_name4', 'new_name4')]
                                   }

        Parameters
        ----------
        other : Mapping[int, list[tuple[str, str]]]
            A mapping of .

        Returns
        -------
        ProcessJoined
            A joined processes.

        Raises
        ------
        ValueError
            raises when the 'other' argument is not a type of a mapping.
        """
        if isinstance(other, Mapping):
            assert len(other) <= len(self.processes), "The number of renaming"
            " must be less than or equal to the number of processes."
            return ProcessJoined(self, kwargs_mapping=other)
        else:
            raise ValueError(
                f"The {type(other)=} must be a Mapping[str, int] ")


class ImageProcessingPipeline(AbstractPipeline):
    def __init__(self, processes=[]):
        super(ImageProcessingPipeline, self).__init__(processes)

    def __rshift__(self, other) -> AbstractPipeline:
        """Appends the process to the end of an Process or ImageProcessingPipeline.

           This is an imumutable operation and new ImageProcessingPipeline will be
           returned.

           Example:
                proc1 >> proc2 >> proc3

        Parameters
        ----------
        other : AbstractProcess | ImageProcessingPipeline
            The process or pipeline that will be appended to the given
            pipeline.
            For ProcessFork in the RHS, the output of the ProcessFork
            will be joined by returning a ProcessJoined object inside a
            pipeline.

        Returns
        -------
        AbstractPipeline
            A new (immutable) pipeline object that contains the current
            pipeline followed by 'other' argument.

        Raises
        ------
        ValueError
            Arises when the RHS is not one of the accptable types.
        """
        if issubclass(type(other), AbstractProcess):
            if isinstance(other, ProcessFork):
                other = ProcessJoined(other)
            return ImageProcessingPipeline(self.processes + [other])
        elif isinstance(other, ImageProcessingPipeline):
            return ImageProcessingPipeline(self.processes + other.processes)
        else:
            raise ValueError(
                f"The '{type(other)}' must be a Process or ImageProcessingPipeline.")

    def __mul__(self, other) -> ProcessFork:
        """Fork two or more pipslines.

           This is an imumutable operation and new ProcessFork will be
           returned.

           Example:
               p1 * p2 : Forks two processes/pipelines
               p1 * p2 * p3 : Forks three processes/piplines

        Parameters
        ----------
        other : int, ProcessFork, AbstractProcess, ImageProcessingPipeline
            When the RHS is n (int), it is forked to n replicate.
            When the RHS is ProcessFork, AbstractProcess or ImageProcessingPipeline,
            the LHS process stack on its process.

        Returns
        -------
        ProcessFork
            A callable object that forks the processes.

        Raises
        ------
        ValueError
            Raises when the 'other' argument is not an int, a ProcessFork, sub class of
            AbstractProcess, or ImageProcessingPipeline.
            Also, in case for RHS int and LHS ProcessFork, it will raise the error.
        """
        if isinstance(other, int):
            return ProcessFork([self] * other)
        if isinstance(other, ProcessFork):
            raise ValueError(
                f"The {type(other)=} cannot be mutiplied from LHS of a ImageProcessingPipeline."
            )
        if issubclass(type(other), AbstractProcess):
            return ProcessFork([self, other])
        if issubclass(type(other), ImageProcessingPipeline):
            return ProcessFork([self, other])
        else:
            raise ValueError(
                f"The {type(other)=} must be an int, a Process, ImageProcessingPipeline "
                f"or tuples of both.")

    def process(self, /, **kwargs) -> Dict:
        """Proccess and return the pipeline.

        all kwargs are default values for any processes in down-stream. However,
        the process returned parameters have precedence.

        Returns
        -------
        Dict
            A dict object that conains the processed image and other parameters.

        Raises
        ------
        IncompatibleArgsException
        """
        try:
            payload_kwargs = Dict(**kwargs)
            for index, process in enumerate(self.processes):
                ret: Dict = process(**payload_kwargs)
                # Unino the returned payload with previous ones.
                # This will be passed to next process or return to the caller.
                #
                #  1- Process parameters have precedence over the kwargs.
                #  2- The latest process parameters have precedence over the formeres.
                payload_kwargs |= ret
        except TypeError as e:
            # The name of the class
            def name(obj) -> str: return type(obj).__name__
            if (len(e.args) > 0 and
                    ("missing 1 required keyword-only argument" in e.args[0] or
                     "missing 1 required positional argument" in e.args[0]
                     )):
                if index > 0:
                    previous_process = name(self.processes[index - 1])
                else:
                    previous_process = "(input of the pipline)"
                raise IncompatibleArgsException(
                    f"The process '{name(process)}' received incompatible payload from "
                    f"the previous process '{previous_process}'.")
            else:
                raise e
        return payload_kwargs


class ProcessLogic(AbstractProcess):
    def __init__(self, logic_callback: Callable[..., Dict]):
        self.logic_callback = logic_callback

    def __call__(self, **kwargs) -> Dict:
        return self.logic_callback(**kwargs)


class ProcessLogicProperty(AbstractProcess):
    def __init__(self, logic_callback: Callable[..., Dict]):
        self.logic_callback = logic_callback
        self.caller_class = None

    def __call__(self, **kwargs) -> Dict:
        return self.logic_callback(self.caller_class, **kwargs)


class ProcessFactory:
    def __init__(self, factory_callback: Callable[..., AbstractProcess]):
        self.factory_callback = factory_callback

    def __call__(self, *args: Any, **kwargs) -> AbstractProcess:
        return self.create(*args, **kwargs)

    def create(self, *args: Any, **kwargs) -> AbstractProcess:
        """Create a parametrised process (arguments) to be attached and called later."""
        process = self.factory_callback(*args, **kwargs)
        return process



import pytest

from pyMOSF.config import Dict, GUIFramework
from pyMOSF.core import (
    AbstractApp,
    AbstractLayout,
    AsyncService,
    Event,
    EventDispatcher,
    EventType,
    ServiceArgumentError,
    ServiceRegistry,
    SyncService,
)

ServiceRegistry._framework = GUIFramework.TOGA


class SingletonSyncService(SyncService):

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SingletonSyncService, cls).__new__(cls)
            # Put any initialization here.
            cls._instance.called = False
            cls._instance.exited = False
        return cls._instance

    def handle_event(self, widget, app, service_callback=None, *args, **kwargs):
        self.called = True
        if service_callback:
            service_callback(Dict(**kwargs))

    def on_exit(self):
        self.exited = True


class SingletonAsyncService(AsyncService):
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SingletonAsyncService, cls).__new__(cls)
            # Put any initialization here.
            cls._instance.called = False
            cls._instance.exited = False
        return cls._instance

    async def handle_event(self, widget, app, service_callback=None, *args, **kwargs):
        self.called = True

    def on_exit(self):
        self.exited = True


class UIElement:
    def __init__(self):
        self.id = "test"
        self.children = []

    def on_press(self):
        pass


class ConfigurableLayout(AbstractLayout):
    def __init__(self, app: AbstractApp) -> None:
        super().__init__(app)
        self.children = []
        self.common_config_called = False
        self.linux_config_called = False

    def build_layout(self, app):
        element = UIElement()
        self.children.append(element)
        return self

    def on_common_config(self):
        self.common_config_called = True
        #
        service = SingletonSyncService()
        event = Event(id="test", eventType=EventType.ON_PRESS,
                      service=service)
        registry = ServiceRegistry()
        registry.bind_event(event)

    def on_linux_config(self):
        self.linux_config_called = True


class ServicableApp(AbstractApp):
    def __init__(self):
        # Set the layout here
        super().__init__(ConfigurableLayout(self))


class FarmeworkApp(AbstractApp):
    def __init__(self):
        # Set the layout here
        super().__init__(ConfigurableLayout(self))
        self.common_config_called = False
        self.linux_config_called = False

    def framework_startup(self):
        """Assumed this function is called by framework"""
        self._set_config()
        self.on_load()

    def on_common_config(self):
        self.common_config_called = True

    def on_linux_config(self):
        self.linux_config_called = True


class LayoutWithCallback(AbstractLayout):
    def __init__(self, app: AbstractApp) -> None:
        super().__init__(app)
        self.children = []
        self.element1 = None
        self.element2 = None
        self.element3 = None
        self.element1_callback_called = False
        self.element2_callback_called = False
        self.element3_callback_called = False
        self.element3_callback_has_results = False
        self.counter = 0
        self.recived_counter = 0

    def build_layout(self, app):
        self.element1 = UIElement()
        self.element1.id = "element1"
        self.children.append(self.element1)
        self.element2 = UIElement()
        self.element2.id = "element2"
        self.children.append(self.element2)
        self.element3 = UIElement()
        self.element3.id = "element3"
        self.children.append(self.element3)
        self.element4 = UIElement()
        self.element4.id = "element4"
        self.children.append(self.element4)
        return self

    def element1_callback(self, results):
        self.element1_callback_called = True

    def element3_callback(self, results):
        self.element3_callback_called = True
        if results.has_results:
            self.element3_callback_has_results = True

    def lazy_value_loading(self):
        self.counter += 1
        return self.counter

    def element4_callback(self, results):
        if results.counter:
            self.recived_counter = results.counter

    def on_common_config(self):
        def local_callback(results):
            self.element2_callback_called = True
        #
        registry = ServiceRegistry()
        service = SingletonSyncService()
        # Bind event for element1
        event = Event(id="element1", eventType=EventType.ON_PRESS,
                      service=service,
                      service_callback=self.element1_callback)
        registry.bind_event(event)
        # Bind event for element2
        event = Event(id="element2", eventType=EventType.ON_PRESS,
                      service=service,
                      service_callback=local_callback)
        registry.bind_event(event)
        # Bind event for element3
        event = Event(id="element3", eventType=EventType.ON_PRESS,
                      service=service,
                      service_callback=self.element3_callback,
                      extra_kwargs={"has_results": True}
                      )
        registry.bind_event(event)
        # Bind event for element4
        event = Event(id="element4", eventType=EventType.ON_PRESS,
                      service=service,
                      service_callback=self.element4_callback,
                      extra_kwargs={
                          "counter": self.lazy_value_loading}
                      )
        registry.bind_event(event)


class LayoutWithCallApp(AbstractApp):
    def __init__(self):
        # Set the layout here
        super().__init__(LayoutWithCallback(self))


@pytest.fixture
def dispatcher():
    return EventDispatcher()


def test_service_registry_register():
    servicable_app = ServicableApp()
    servicable_app.on_load()  # Bind event dispatcher will happen here
    element = servicable_app.layout.children[0]  # type: ignore
    assert callable(element.on_press)
    element.on_press()
    service = SingletonSyncService()
    assert service.called


def test_service_registry_callbacks():
    #
    layout_callback_app = LayoutWithCallApp()
    layout_callback_app.on_load()  # Bind event dispatcher will happen here
    #
    element = layout_callback_app.layout.element1  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.element1_callback_called  # type: ignore
    #
    element = layout_callback_app.layout.element2  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.element2_callback_called  # type: ignore
    #
    element = layout_callback_app.layout.element3  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.element3_callback_called  # type: ignore
    assert layout_callback_app.layout.element3_callback_has_results  # type: ignore
    #
    element = layout_callback_app.layout.element4  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.recived_counter == 1  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.recived_counter == 2  # type: ignore
    element.on_press()
    assert layout_callback_app.layout.recived_counter == 3  # type: ignore


def test_event_dispatcher_dispatch(dispatcher):
    called = False

    def handler(widget, app, callback, **kwargs):
        nonlocal called
        called = True
    dispatcher.register("event1", handler)
    dispatcher.dispatch("event1", None, None, None, None)  # type: ignore
    assert called


def test_event_dispatcher_callback(dispatcher):
    called = False

    def callback(widget, app, **kwargs):
        nonlocal called
        called = True

    def handler(widget, app, callback, **kwargs):
        callback(widget, app)
    dispatcher.register("event1", handler)
    dispatcher.dispatch("event1", None, None, callback, None)  # type: ignore
    assert called


def test_service_registry_on_exit():
    service = SingletonSyncService()
    event = Event(id="exit_test", eventType=EventType.ON_PRESS,
                  service=service)
    registry = ServiceRegistry()
    registry.bind_event(event)
    registry.on_exit(None)
    assert service.exited


# @pytest.mark.asyncio
# async def test_event_dispatcher_dispatch_async(dispatcher):
#     called = False

#     async def async_handler(widget, app, **kwargs):
#         nonlocal called
#         called = True
#     dispatcher.register_async("async_event", async_handler)
#     await dispatcher.dispatch_async("async_event", None, None, None, None)
#     assert called


def test_service_argument_error():
    with pytest.raises(ServiceArgumentError) as excinfo:
        raise ServiceArgumentError("Test message")
    assert str(excinfo.value) == "Test message"


def test_configurable():
    servicable_app = ServicableApp()
    servicable_app.on_load()  # Bind event dispatcher will happen here
    assert servicable_app.layout.common_config_called   # type: ignore
    assert servicable_app.layout.linux_config_called   # type: ignore


def test_app_configurable():
    framework_app = FarmeworkApp()
    framework_app.framework_startup()  # It is called by framework
    assert framework_app.common_config_called   # type: ignore
    assert framework_app.linux_config_called   # type: ignore
    # Layout should be configured too
    assert framework_app.layout.common_config_called   # type: ignore
    assert framework_app.layout.linux_config_called   # type: ignore



"""
 Life-cycle sequence:
1- KivyMultiLayoutApp.__init__ : manager, _layout_lookup, _screen_lookup
    1-1- AbstractApp.__init__ : _settings, layout

2- KivyStackedLayout.__init__  : types, _instances
    2-1- KivyLayout.__init__ : _main_box
        2-1-1- AbstractLayout.__init__ : _app


3-KivyMultiLayoutApp.build : main_window
    3-1- main_window.add_widget(self._current_layout.main_box)
    3-2- manager.add_widget(self.main_window)
    3-3- AbstractApp.on_begin :
        3-3-1- AbstractApp._set_config :
        3-3-2- AbstractApp.on_load :
            3-3-2-1- AbstractLayout.build_layout -> root :
                3-3-2-1-1- Layout[Box]._build_box -> _box
                    3-3-2-1-1-1- Layout[Box]._build_box
                        3-3-2-1-1-1-1- KivyLayout._build_box :  create box and return

                    or alternativly

                    3-3-2-1-1-1- StackedLayout[Box, *Ts].build_layout
                        3-3-2-1-1-1-1 StackedLayout[Box, *Ts]._build_box
                            3-3-2-1-1-1-1-1 Component[Box, *Ts]._init_
                                3-3-2-1-1-1-1-1-1- KivyComponent._init_: layout, add ids
                        3-3-2-1-1-1-2 self._add_to_main_box(self._box, component)
                        3-3-2-1-1-1-3 _get_component_passthrough
                        3-3-2-1-1-1-4 Component[Box, *Ts].on_load

            3-3-2-2- AbstractApp.__enumerate_elements(root) :
            3-3-2-3- AbstractLayout._set_config :
            3-3-2-4- Register Services of 'root'
            3-3-2-5- AbstractLayout.on_load :
    3-4- self._add_layout(self._current_layout)

"""

from __future__ import annotations

import logging
from pathlib import Path

from kivy.app import App as kivyApp  # type: ignore #

# from kivy.core.window import Window  # type: ignore
from kivy.uix.boxlayout import BoxLayout  # type: ignore
from kivy.uix.screenmanager import Screen, ScreenManager  # type: ignore

from pyMOSF.config import Configurable, GUIFramework, Settings
from pyMOSF.core import (
    Component,
    Layout,
    MultiLayoutApp,
    ServiceRegistry,
    StackedLayout,
)

ServiceRegistry._framework = GUIFramework.KIVY

log = logging.getLogger(__name__)


class KivyLayout(Layout):
    def __init__(self, app: KivyMultiLayoutApp) -> None:
        super().__init__(app)
        self._app = app

    @property
    def ids(self):
        if self._box is None:
            raise ValueError("box has not been initialized (_build_box).")
        return self._box.ids

    @property
    def ml_app(self) -> KivyMultiLayoutApp:
        return self._app


class KivyBox(BoxLayout):
    """This class is used for element inkv files"""

    def __init__(self, layout: KivyLayout, **kwargs) -> None:
        super().__init__(**kwargs)
        self._layout = layout
        for key in self.ids:
            setattr(self.ids[key], "id", key)

    @property
    def ml_app(self) -> KivyMultiLayoutApp:
        return self._layout.ml_app

    @property
    def parrent_layout(self) -> KivyLayout:
        return self._layout


class KivyComponent(Component, Configurable, BoxLayout):
    def __init__(self, layout: KivyStackedLayout, **kwargs) -> None:
        super().__init__(layout, **kwargs)
        self._layout = layout
        for key in self.ids:
            setattr(self.ids[key], "id", key)

    @property
    def ml_app(self) -> KivyMultiLayoutApp:
        return self._layout.ml_app

    @property
    def parrent_layout(self) -> KivyStackedLayout:
        return self._layout


class KivyStackedLayout(StackedLayout):

    def __init__(self,
                 app: KivyMultiLayoutApp,
                 *types):  # type: ignore
        self._main_container = BoxLayout()
        self._main_container.orientation = 'vertical'
        super().__init__(app, self._main_container, *types)

    @property
    def ml_app(self) -> KivyMultiLayoutApp:
        if self._app is None:
            raise ValueError("app has not been initialized.")
        return self._app

    @ml_app.setter
    def ml_app(self, value: KivyMultiLayoutApp):
        self._app = value

    def _add_to_main_box(self, box: BoxLayout, element: BoxLayout):
        box.add_widget(element)


class KivyMultiLayoutApp(MultiLayoutApp, kivyApp):
    def __init__(self,
                 init_layout: KivyLayout | KivyStackedLayout,
                 settings: Settings | None = None,
                 **kwargs) -> None:
        self._main_container = BoxLayout()
        self._main_container.orientation = 'vertical'
        self.manager = ScreenManager()
        self._screen_lookup = {}
        self._current_layout = init_layout
        super().__init__(self._main_container, init_layout, settings, **kwargs)

    @property
    def path(self):
        return Path(self.directory)  # type: ignore

    @property
    def data_path(self):
        return Path(self.user_data_dir)  # type: ignore

    @property
    def current_layout(self) -> KivyLayout | KivyStackedLayout:
        return self._current_layout

    def _add_layout(self, layout: KivyLayout | KivyStackedLayout):
        if layout._name not in self._screen_lookup:
            self.main_window: Screen = Screen(name=layout._name)
            self._screen_lookup[layout._name] = self.main_window
            self.main_window.add_widget(self.layout.main_box)
            self.manager.add_widget(self.main_window)
            self.manager.current = layout._name
        else:
            self.manager.current = layout._name

    def build(self):
        """Construct and show the Kivy application (it's called by kivy app).

        It also call the layout build and bind the serivices handlers
        to UI elements on_press call back, if their ids correspond to
        an id in ServiceRegistry.
        """
        self.main_window: Screen = Screen(name=self.layout._name)
        self._screen_lookup[self.layout._name] = self.main_window
        self.main_window.add_widget(self._current_layout.main_box)
        self.manager.add_widget(self.main_window)
        # Call on_load in AbstractApp to register the event once
        self.on_begin()
        # It calls once
        self._add_layout(self._current_layout)
        #
        return self.manager

    def on_stop(self):
        """Alrert layout and its components the end of app cycle (it's called by kivy app)."
        """
        self.on_end()  # self.layout.on_end() will be called here
        return super().on_stop()



import logging
import os

import kivy
from kivy.lang import Builder  # type: ignore
from kivy.logger import Logger  # type: ignore

from pyMOSF.kivy import KivyMultiLayoutApp  # , Layout
from pyMOSF.templates.kivy.example_layouts import ExampleViewLayout

logging.root = Logger

kivy.require('1.8.0')  # type: ignore
log = logging.getLogger(__name__)


class KivyApp(KivyMultiLayoutApp):
    def __init__(self, **kwargs):
        #
        self.main_layout = ExampleViewLayout(self)
        super(KivyApp, self).__init__(
            init_layout=self.main_layout,
            **kwargs)
        self.layout = self.main_layout

    def show_main(self):
        self.show_layout(self.main_layout)


def create_app():
    log.info("Example app stars")
    app = KivyApp()
    return app


if __name__ == '__main__':
    # This main entry point is used for debugging.
    # It is not used in the main app.
    log.info("Debug main is called.")
    Builder.load_file(os.path.abspath(
        "pyMOSF/templates/kivy/example_layouts.kv"))
    log.info("KV is loaded.")
    create_app().run()



import logging
import os
from pathlib import Path
from typing import Any, Callable

from kivy.factory import Factory  # type: ignore
from kivy.lang import Builder  # type: ignore

# import toga
from kivy.properties import ObjectProperty  # type: ignore
from kivy.uix.floatlayout import FloatLayout  # type: ignore
from kivy.uix.popup import Popup  # type: ignore

import pyMOSF.core as core
from pyMOSF.config import Dict
from pyMOSF.core import safe_call

log = logging.getLogger(__name__)


class LoadDialog(FloatLayout):
    load = ObjectProperty(None)
    cancel = ObjectProperty(None)


Factory.register('LoadDialog', cls=LoadDialog)
Builder.load_string('''
<LoadDialog>:
    BoxLayout:
        size: root.size
        pos: root.pos
        orientation: "vertical"
        FileChooserIconView:
            id: filechooser

        BoxLayout:
            size_hint_y: None
            height: 30
            Button:
                text: "Cancel"
                on_release: root.cancel()

            Button:
                text: "Load"
                on_release: root.load(filechooser.path, filechooser.selection)
                    ''')


class FileOpen(core.SyncService):
    def __init__(self,
                 dialog_title: str = "Open file",
                 initial_directory: Path | str | None = None,
                 file_types: list[str] | None = None,
                 multiple_select: bool = False) -> None:
        super().__init__()
        self.app = None
        self.dialog_title = dialog_title
        self.initial_directory = initial_directory
        self.file_types = file_types
        self.multiple_select = multiple_select

    def dismiss_popup(self):
        self._popup.dismiss()

    @safe_call(log)
    def load(self, path, filename):
        if len(filename) > 0:
            if self.service_callback is not None:
                if self.multiple_select:
                    # If multiple files are selected, return a list of file names
                    self.service_callback(Dict(files=filename))
                else:
                    # If only one file is selected, return the file name
                    self.service_callback(Dict(file=filename[0]))
        self._popup.dismiss()

    @safe_call(log)
    def handle_event(self,
                     widget: Any,
                     app: core.AbstractApp,
                     service_callback: Callable | None = None,
                     *args,
                     **kwargs):

        self.app = app
        content = LoadDialog(load=self.load, cancel=self.dismiss_popup)
        if self.initial_directory is None:
            content.ids['filechooser'].path = os.getcwd()
        else:
            content.ids['filechooser'].path = str(self.initial_directory)
        if self.file_types is not None:
            content.ids['filechooser'].filters = self.file_types
        content.ids['filechooser'].multiselect = self.multiple_select
        self._popup = Popup(title=self.dialog_title, content=content,
                            size_hint=(0.9, 0.9))
        self.service_callback = service_callback
        self._popup.open()
        return



import logging
from enum import Enum
from pathlib import Path
from typing import Any

from rubicon.objc import ObjCClass, ObjCProtocol, objc_block, objc_method, py_from_ns
from rubicon.objc.runtime import objc_id

import pyMOSF.core as core
from pyMOSF.core import safe_async_call

log = logging.getLogger(__name__)

# Import necessary UIKit classes
UIImagePickerController = ObjCClass("UIImagePickerController")
UIApplication = ObjCClass("UIApplication")
NSObject = ObjCClass("NSObject")
UIViewController = ObjCClass("UIViewController")
UIImagePickerControllerDelegate = ObjCProtocol(
    "UIImagePickerControllerDelegate")
UINavigationControllerDelegate = ObjCProtocol("UINavigationControllerDelegate")


class ImagePickerReturn(int, Enum):
    All = 0  # returns all the informations from the picker
    PATH = 1  # returns the path of the selected file
    IMAGE = 2  # returns the image data


class PhotoPickerDelegate(NSObject,  # type: ignore
                          protocols=[UIImagePickerControllerDelegate,  # type: ignore
                                     UINavigationControllerDelegate]):  # type: ignore

    @objc_method
    def init_(self):
        self = self.init()
        if self is None:
            return None
        self.serviceCallback = None
        return self

    @objc_method
    def set_serviceCallback(self, serviceCallback: objc_block):
        self.serviceCallback = serviceCallback

    @objc_method
    def get_serviceCallback(self):
        return self.serviceCallback

    @objc_method
    def imagePickerController_didFinishPickingMediaWithInfo_(self,
                                                             picker: objc_id,
                                                             info: objc_id) -> None:
        # Handle the selected photo
        if self.serviceCallback is not None:
            self.serviceCallback(info)  # type: ignore
        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.dismissViewControllerAnimated_completion_(
            True, None)

    @objc_method
    def imagePickerControllerDidCancel_(self, picker: objc_id) -> None:
        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.dismissViewControllerAnimated_completion_(
            True, None)


class IOSPhotoPicker(core.AsyncService):
    def __init__(self, imagePickerReturn: ImagePickerReturn = ImagePickerReturn.IMAGE) -> None:
        super().__init__()
        # IMPORTANT: the delegate must be created and stored here
        # because the delegate is a weak reference
        # and can be deleted without any reference in Python side
        self.delegate = PhotoPickerDelegate.alloc().init_()
        self.imagePickerReturn = imagePickerReturn
        self.picker = None

    @safe_async_call(log)
    async def handle_event(self,
                           widget: Any,
                           app,
                           service_callback,
                           *args, **kwargs):
        if service_callback is None:
            raise ValueError("service_callback must be provided")

        self.picker = UIImagePickerController.alloc().init()  # type: ignore
        self.picker.sourceType = 0  # 0 = Photo Library

        def local_callback(info_c: objc_id) -> None:
            # Convert the NSDictionary to a Python dictionary
            info_dict: dict | Any = py_from_ns(info_c)
            photo_dict = {}
            info_dict: dict | Any = py_from_ns(info_c)

            match self.imagePickerReturn:
                case ImagePickerReturn.All:
                    photo_dict = {}
                    for key, value in info_dict.items():
                        key_str: str | Any = py_from_ns(key)
                        photo_dict[key_str] = py_from_ns(value)
                    service_callback(photo_dict)
                    return
                case ImagePickerReturn.PATH:
                    for key, value in info_dict.items():
                        if py_from_ns(key) == "UIImagePickerControllerImageURL":
                            path_str: str = py_from_ns(
                                value.path)  # type: ignore
                            service_callback(Path(path_str))
                            return
                case ImagePickerReturn.IMAGE:
                    path_str = ""
                    for key, value in info_dict.items():
                        if py_from_ns(key) == "UIImagePickerControllerImageURL":
                            path_str: str = py_from_ns(
                                value.path)  # type: ignore
                    if path_str != "":
                        from PIL import Image

                        # file.startAccessingSecurityScopedResource()
                        img = Image.open(Path(path_str))
                        # file.stopAccessingSecurityScopedResource()
                        service_callback(img)
                        return

                    # Set the service callback
        self.delegate.set_serviceCallback(local_callback)  # type: ignore
        # Set the delegate
        self.picker.delegate = self.delegate
        # Present the photo picker
        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.presentViewController_animated_completion_(
            self.picker, True, None)




def int_(value, /, base: int = 10, default: int = 0):
    try:
        return int(value)
    except TypeError:
        return default


def __chain_traceback(tb, message: str = "", counter: int = 1) -> str:
    """A recursive call that extract the exceptions' files.

    Parameters
    ----------
    tb : TracebackType
        exceptions's __traceback__.
    message : str, optional
        The chain of file paths and line numbers, by default ""
    counter : int, optional
        The chain order number, by default 1
    Returns
    -------
    str
       The chain of file paths and line numbers from exceptions's __traceback__.
    """
    if tb is None:
        return message
    #
    if tb.tb_next is not None:
        message += __chain_traceback(tb.tb_next, message, counter+1)
    frame = tb.tb_frame
    message += f'\n{counter}↦ "{frame.f_code.co_filename}", line {frame.f_lineno}'
    return message


def safe_call(log, exceptions={}):
    """A decorator for safely calling a sync function and logging the errors.

    Parameters
    ----------
    log : logging.Logger
        A logger instance to write the errors.
    exceptions : dict, optional
        Dictionary of exceptions and thier corresponding messages to catch and
        log, instead of showing file and line number, by default {}
    """
    def try_call(func):

        def caller(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if type(e) in exceptions:
                    log.error(f"{exceptions[type(e)]}")
                else:
                    log.error(
                        f"\n{type(e).__name__}:{e}\n{__chain_traceback(e.__traceback__)}\n")
        return caller
    return try_call


def safe_async_call(log, exceptions={}):
    """A decorator for safely calling a async function and logging the errors.

    Parameters
    ----------
    log : logging.Logger
        A logger instance to write the errors.
    exceptions : dict, optional
        Dictionary of exceptions and thier corresponding messages to catch and
        log, instead of showing file and line number, by default {}
    """
    def try_call(func):
        async def caller(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                if type(e) in exceptions:
                    log.error(f"{exceptions[type(e)]}")
                else:
                    log.error(
                        f"\n{type(e).__name__}:{e}\n{__chain_traceback(e.__traceback__)}\n")
        return caller
    return try_call



import logging
import shutil
from pathlib import Path
from typing import Any

from rubicon.objc import (
    ObjCClass,
    ObjCProtocol,
    objc_block,
    objc_const,
    objc_method,
    py_from_ns,
)
from rubicon.objc.runtime import Foundation, load_library, objc_id

import pyMOSF.core as core
from pyMOSF.config import Dict
from pyMOSF.core import safe_async_call

log = logging.getLogger(__name__)

# Import necessary UIKit classes
UIDocumentPickerViewController = ObjCClass("UIDocumentPickerViewController")
UIApplication = ObjCClass("UIApplication")
NSObject = ObjCClass("NSObject")
UIViewController = ObjCClass("UIViewController")
UIDocumentPickerDelegate = ObjCProtocol("UIDocumentPickerDelegate")
NSURL = ObjCClass("NSURL")
NSString = ObjCClass("NSString")
NSNotificationCenter = ObjCClass('NSNotificationCenter')
NSLOG = Foundation.NSLog

NSFileManager = ObjCClass('NSFileManager')


class DocumentPickerDelegate(NSObject,  # type: ignore
                             protocols=[UIDocumentPickerDelegate]):  # type: ignore

    @objc_method
    def init_(self):
        self = self.init()
        if self is None:
            return None
        self.serviceCallback = None
        return self

    @objc_method
    def set_serviceCallback(self, serviceCallback: objc_block):
        self.serviceCallback = serviceCallback

    @objc_method
    def get_serviceCallback(self):
        return self.serviceCallback

    @objc_method
    def documentPicker_didPickDocumentsAtURLs_(self, picker: objc_id, urls: objc_id) -> None:
        # Handle the selected files
        if self.serviceCallback is not None:
            self.serviceCallback(urls)  # type: ignore

        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.dismissViewControllerAnimated_completion_(
            True, None)

    @objc_method
    def documentPickerWasCancelled_(self, picker: objc_id) -> None:
        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.dismissViewControllerAnimated_completion_(
            True, None)


class IOSFileOpen(core.AsyncService):
    def __init__(self,
                 initial_directory: Path,
                 file_types: list[str] = ["UTTypeData"],
                 multiple_select: bool = True,
                 ) -> None:
        super().__init__()
        # IMPORTANT: the delegate must be create and store here
        # because the delegate is a weak reference
        # and can be deleted without any reference in Python side
        self.delegate = DocumentPickerDelegate.alloc().init_()
        self.temp_path = initial_directory
        self.document_types = file_types
        self.allowsMultipleSelection = multiple_select
        self.libcf = load_library("UniformTypeIdentifiers")
        self.picker = None

    @safe_async_call(log)
    async def handle_event(self,
                           widget: Any,
                           app,
                           service_callback,
                           *args, **kwargs):
        if service_callback is None:
            raise ValueError("service_callback must be provided")

        # You can specify other UTIs if needed
        document_types = [objc_const(self.libcf, item)
                          for item in self.document_types]
        self.picker = UIDocumentPickerViewController.alloc()  # type: ignore
        self.picker = self.picker.initForOpeningContentTypes_asCopy_(
            document_types, True)
        self.picker.allowsMultipleSelection = self.allowsMultipleSelection

        def local_callback(fnames_c: objc_id) -> None:
            # convert the NSArray to a Python list
            fnames_c_list = py_from_ns(fnames_c)
            fnames = []
            for item in fnames_c_list:  # type: ignore
                # All files are copied to local sandbox , before
                # the service caller's callback is called.
                # Therefore, all paths are updated accordingly
                # This steps must happen between a SecurityScope block
                item.startAccessingSecurityScopedResource()
                path_str: str = py_from_ns(item.path)  # type: ignore
                f_src = open(path_str, 'rb')
                dest_path = self.temp_path / Path(path_str).name
                f_dest = open(dest_path, 'wb')
                shutil.copyfileobj(f_src, f_dest)
                item.stopAccessingSecurityScopedResource()
                # store the converted Path object
                fnames.append(dest_path)

            if self.allowsMultipleSelection:
                service_callback(Dict(files=fnames))
            else:
                if len(fnames) > 0:
                    service_callback(Dict(file=fnames[0]))
        # Set the service callback
        self.delegate.set_serviceCallback(local_callback)  # type: ignore
        # Set the delegate
        self.picker.delegate = self.delegate
        # Present the document picker
        keyWindow = UIApplication.sharedApplication.keyWindow  # type: ignore
        keyWindow.rootViewController.presentViewController_animated_completion_(
            self.picker, True,  None)



from setuptools import setup

if __name__ == "__main__":
    # # Detect which extras are being installed
    # install_kivy = any('kivy' in arg for arg in sys.argv)
    # install_beeware = any('beeware' in arg for arg in sys.argv)

    # # Base packages that are always included
    # packages = find_packages(include=['pyMOSF',
    #                                   'pyMOSF.config*',
    #                                   'pyMOSF.services*',
    #                                   'pyMOSF.core*'])

    # # Add framework-specific packages based on extras
    # if install_kivy:
    #     packages.extend(find_packages(include=['pyMOSF.kivy*']))
    # if install_beeware:
    #     packages.extend(find_packages(include=['pyMOSF.toga*']))

    # # Override the packages option from setup.cfg
    # setup(packages=packages)
    setup()



import logging
import os
from pathlib import Path

is_not_kivy = True
try:
    from kivy.logger import Logger  # type: ignore
    logging.root = Logger
    root = logging.getLogger().root
    is_not_kivy = False
except ImportError:
    root = logging.getLogger().root


class OneLineExceptionFormatter(logging.Formatter):
    def formatException(self, ei):
        result = super().formatException(ei)
        return repr(result)

    def format(self, record):
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        return result


def std_out_log(log_level="INFO",
                LOGGER_FORMAT=logging.BASIC_FORMAT):
    handler = logging.StreamHandler()
    # LOGGER_FORMAT = "%(levelname)s:%(name)s:%(message)s \n %(pathname)s @ %(lineno)d"
    # LOGGER_FORMAT = "%(levelname)s:%(name)s:%(message)s (in %(filename)s @ line %(lineno)d)"
    formatter = OneLineExceptionFormatter(LOGGER_FORMAT)
    handler.setFormatter(formatter)
    # root = logging.getLogger()
    root.setLevel(os.environ.get("LOGLEVEL", log_level))
    if is_not_kivy:
        root.addHandler(handler)


def file_out_log(path,
                 log_level="INFO",
                 LOGGER_FORMAT="%(levelname)s:%(asctime)s:%(name)s:%(message)s"):
    import logging.handlers

    # from logging.handlers import RotatingFileHandler
    # handler = RotatingFileHandler(os.environ.get("LOGFILE", path),
    #                               maxBytes=10000,
    #                               backupCount=3)
    if not os.path.exists(path):
        try:
            file_path = os.environ.get("LOGFILE", path)
            with open(file_path, 'w') as file:
                file.write("")
        except FileNotFoundError:
            file_path = os.environ.get("LOGFILE", path)
            directories, _ = os.path.split(file_path)
            Path(directories).mkdir(parents=True, exist_ok=True)
    handler = logging.handlers.WatchedFileHandler(
        os.environ.get("LOGFILE", path))
    # LOGGER_FORMAT = logging.BASIC_FORMAT
    # LOGGER_FORMAT = "%(levelname)s:%(name)s:%(message)s \n %(pathname)s @ %(lineno)d"
    # LOGGER_FORMAT = "%(levelname)s:%(name)s:%(message)s (in %(filename)s @ line %(lineno)d)"
    formatter = logging.Formatter(LOGGER_FORMAT)
    handler.setFormatter(formatter)
    # root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(os.environ.get("LOGLEVEL", log_level))


def __dummy__():
    pass


if __name__ == "__main__":
    std_out_log()



import functools
from typing import Any, Callable

from pyMOSF.config import Dict
from pyMOSF.core import EventType
from pyMOSF.core.pipelines import (
    AbstractProcess,
    ProcessFactory,
    ProcessLogic,
    ProcessLogicProperty,
)


def component_method(func: Callable) -> Callable[..., Any]:
    """A decorator for making component's method available to layout.

       It do nothing and just annotate a class function.

       Example:
       --------
       class ExampleComponent(TogaComponent):
           @component_method
           def is_bw(self):
               return self.swt_black_white.value

           @property
           def is_gray(self):
               return self.swt_gray.value

        then, the layout that contains the 'ExampleComponent'
        can access 'is_bw' function simply like self.is_bw()

        However, for properties, you need to use the getitem
        self[ExampleComponent].is_gray

    """
    return func


# def service_callback(func):
#     """Decorate the method that as a callback that can be set as an argument.
#     """
#     def callable_func(*args_local):
#         if len(args_local) == 0:  # ordinary function
#             return func

#         def wraper(*args, **kwargs):  # object methods
#             func(args_local[0], *args, **kwargs)
#         return wraper
#     return callable_func


def silence_crossed_events(eventType: EventType, element1,  element2):
    """Silence the circular events calls between two elements.

       When a value of the second element is changed by event handler of the first one,
       such that this change fires the same event in the second one, we get a circular
       event from the second. This method let us to disble it.
    """
    _silence(element1, eventType, element2)
    _silence(element2, eventType, element1)


def _silence(element, eventType: EventType, second_element):
    try:
        import toga  # noqa
    except ImportError:
        raise NotImplementedError("This function is only available for Toga.")
    eventName = eventType.name.lower()
    # The element's event that is supposed to be called
    event = getattr(element, f"_{eventName}")
    # Decorator for the element's event that is supposed to be called

    def decorator(widget,  *args):
        # Temporarily store and disable the second element's event
        stored = getattr(second_element, f"_{eventName}")
        setattr(second_element, eventName, None)
        # call the original event handler
        event(widget,  *args)
        # Set back the second element's event
        setattr(second_element, eventName, stored)
    # reroute the element's event handler to the decorator
    setattr(element, eventName, decorator)


def processLogic(func: Callable[..., Dict]) -> Callable[..., ProcessLogic]:
    """A decorator for creating inline process.

       It can turn a function to a process, as long as the function
       arguments include '**kwargs' and returns a Dict as it payloads.
       In short, decorating a function can turn it into an inline definition
       of a process, which can be joined (>>) or forked (*) by other processes.

       Example:
           @processLogic
           def proc2(condition:bool, **kwargs) -> Dict:
               if condition:
                   return Dict(value=True)
               else:
                   return Dict(value=False)

            proc1 >> proc2()

    Parameters
    ----------
    func : Callable[..., Dict]
        Callable function that contains the logic of creating the process.
        The Callable function must include '**kwargs' in its arguments and
        returns a Dict as it payloads.

    Returns
    -------
    Callable[..., ProcessLogic]
        The decorated process. By calling it without any argument,
        it returns a process that can be joined or forked to others.
    """
    p_logic = ProcessLogic(func)

    def logic() -> ProcessLogic:
        return p_logic
    return logic


def processLogicProperty(func: Callable[..., Dict]) -> property:
    """A decorator for creating inline process inside classes as a property.

       It can turn a class method to a process as a property, as long as the method
       arguments includes '**kwargs' and returns a Dict as it payloads.
       In short, decorating a class method turns it into an inline definition
       of a process, which can be joined (>>) or forked (*) by other processes.

       Example:
           class ClassExample:
               @processLogicProperty
               def proc2(self, condition:bool, **kwargs) -> Dict:
                   if condition:
                       return Dict(value=True)
                   else:
                       return Dict(value=False)
                def another _function(self):
                    pipline = proc1 >> self.proc2

    Parameters
    ----------
    func : Callable[..., Dict]
        A method that defined in a class that contains the logic of creating
        the process. The Callable method must have '**kwargs' in its arguments
        (including the default 'self' argument) and returns a Dict as it payloads.

    Returns
    -------
    Callable[..., ProcessLogic]
        The decorated process. By using it like a property,
        it returns a process that can be joined or forked to others.
    """
    p_logic = ProcessLogicProperty(func)

    def logic(caller) -> ProcessLogicProperty:
        p_logic.caller_class = caller
        return p_logic
    # Turns the return to a property with a getter method
    logic_property = property(fget=logic)
    return logic_property


def processFactory(cache: bool,
                   cache_size: int = 256):
    """A factory decorator that turns a function to a process.

       The function must return a process or ImageProcessingPipeline.

    Parameters
    ----------
    cache : bool
        Cache the created pipeline based on the arguments that pass to
        the factory function.
    cache_size : int, optional
        The size of the cache, by default 256
    """
    def decorator(func):
        p_factory = ProcessFactory(func)

        if cache:
            @functools.lru_cache(cache_size)
            def cached_factory(*args, **kwargs) -> AbstractProcess:
                """Parametrisation arguments are passed here."""
                return p_factory.create(*args, **kwargs)
            return cached_factory
        else:
            def factory(*args, **kwargs) -> AbstractProcess:
                """Parametrisation arguments are passed here."""
                return p_factory.create(*args, **kwargs)
            return factory
    return decorator
